package com.intellij.hub.auth.oauth2.token;

/*Generated by MPS */

import com.intellij.hub.auth.AuthToken;
import org.jetbrains.annotations.NotNull;
import java.util.Date;
import org.jetbrains.annotations.Nullable;
import java.util.Arrays;
import javax.xml.bind.DatatypeConverter;
import java.util.regex.Pattern;
import com.intellij.hub.auth.InvalidTokenDataException;
import com.intellij.hub.auth.InvalidTokenSignatureException;
import com.intellij.hub.auth.InvalidTokenException;
import com.intellij.hub.auth.InvalidAuthorizationHeaderException;


/**
 * * @author Danila Ponomarenko
 */
public class AccessToken implements AuthToken {
  public static final String BEARER_TYPE = "Bearer";
  private static final String DATA_DELIMITER = ";";
  private static final String DELIMITER = ".";
  public static final String BEARER_PREFIX = BEARER_TYPE + " ";

  private final AccessTokenData data;
  private final int keyVersion;
  private final byte[] signature;

  public AccessToken(@NotNull AccessTokenData data, int keyVersion, @NotNull byte[] signature) {
    this.data = data;
    this.keyVersion = keyVersion;
    this.signature = signature;
  }

  @NotNull
  public Date getExpirationDate() {
    return data.getExpiration();
  }

  @NotNull
  public int getTTLSeconds() {
    return data.getTTLSeconds();
  }

  @NotNull
  public String getClient() {
    return data.getClient();
  }

  @NotNull
  @Deprecated
  public String getTarget() {
    return getScope()[0];
  }

  @NotNull
  public String[] getScope() {
    return data.getScope();
  }

  @Nullable
  public String getUser() {
    return data.getUser();
  }

  @NotNull
  public AccessTokenData getData() {
    return data;
  }

  public int getKeyVersion() {
    return keyVersion;
  }

  @NotNull
  public byte[] getSignature() {
    return signature;
  }

  @Override
  public boolean equals(Object o) {
    if (this == o) {
      return true;
    }
    if (o == null || getClass() != o.getClass()) {
      return false;
    }
    AccessToken that = (AccessToken) o;
    return data.equals(that.data) && Arrays.equals(signature, that.signature);
  }

  @Override
  public int hashCode() {
    int result = data.hashCode();
    result = 31 * result + Arrays.hashCode(signature);
    return result;
  }

  @Override
  public String toString() {
    return String.format("%s{data=%s, signature=%s}", this.getClass().getSimpleName(), data, DatatypeConverter.printBase64Binary(signature));
  }

  @NotNull
  public String encode() {
    return data.encode() + DATA_DELIMITER + encodeVersionAndSignature();
  }

  @NotNull
  private String encodeVersionAndSignature() {
    return keyVersion + DELIMITER + DatatypeConverter.printBase64Binary(signature);
  }

  @NotNull
  public String getType() {
    return BEARER_TYPE;
  }

  @NotNull
  public static AccessToken decode(@NotNull String token) {
    final String[] parts = token.split(Pattern.quote(DATA_DELIMITER));
    if (parts.length != 2) {
      throw new InvalidTokenDataException("Invalid token data");
    }
    final String[] versionAndSignature = parts[1].split(Pattern.quote(DELIMITER));
    if (versionAndSignature.length != 2) {
      throw new InvalidTokenSignatureException();
    }
    byte[] signature;
    try {
      signature = DatatypeConverter.parseBase64Binary(versionAndSignature[1]);
    } catch (Exception e) {
      throw new InvalidTokenSignatureException(e);
    }
    return new AccessToken(AccessTokenData.decode(parts[0]), Integer.valueOf(versionAndSignature[0]), signature);
  }

  @NotNull
  public static AccessToken decodeHeader(@NotNull String header) throws InvalidTokenException {
    if (!(isBearer(header))) {
      throw new InvalidAuthorizationHeaderException(BEARER_TYPE);
    }
    return AccessToken.decode(header.substring(BEARER_PREFIX.length()));
  }

  @NotNull
  public static String encodeHeader(@NotNull AccessToken token) {
    return BEARER_PREFIX + token.encode();
  }

  public static boolean isBearer(String header) {
    return (header != null && header.length() > 0) && header.startsWith(BEARER_PREFIX);
  }
}
