package jetbrains.jetpass.client.resource;

/*Generated by MPS */

import jetbrains.jetpass.client.accounts.BaseAccountsClient;
import org.jetbrains.annotations.NotNull;
import com.intellij.hub.auth.InvalidTokenException;
import org.jetbrains.annotations.Nullable;
import com.intellij.hub.auth.oauth2.token.AccessToken;
import com.intellij.hub.auth.basic.BasicAuthToken;
import com.intellij.hub.auth.InvalidAuthorizationHeaderException;
import com.intellij.hub.auth.oauth2.provider.AccessTokenVerifier;

public class RequestValidator {
  private final BaseAccountsClient client;
  private volatile String guestId = null;
  private volatile String clientServiceId = null;
  private final DSAPublicKeySource keySource;

  public RequestValidator(@NotNull final BaseAccountsClient client) {
    this.keySource = new DSAPublicKeySource(client.getHubClient());
    this.client = client;
  }

  /**
   * Reads token from header (basic or bearer):
   * if header is not present returns null
   * if header is valid returns user id if token contains user id, otherwise null
   * 
   * @param header Authorization header with token
   * @throws InvalidTokenException if token was invalid or can't be verified
   * @return authenticated user id - if header is valid - returns logged user id, otherwise null (if token doesn't contain user - returns null), if header is empty - returns guest is
   */
  @Nullable
  public String getUserId(String header) throws InvalidTokenException {
    if ((header == null || header.length() == 0)) {
      return null;
    }
    return getAccessToken(header).getUser();
  }

  /**
   * Reads token from header (basic or bearer):
   * if header is not present returns guest id
   * if header is valid returns user id if token contains user id, otherwise null
   * 
   * @param header Authorization header
   * @throws InvalidTokenException if token was invalid or can't be verified
   * @return authenticated user id - if header is valid - returns logged user id, otherwise null (if token doesn't contain user - returns null), if header is empty - returns guest is
   */
  @NotNull
  public String getUserIdOrGuestId(String header) throws InvalidTokenException {
    if ((header == null || header.length() == 0)) {
      return getGuestId();
    }
    return getAccessToken(header).getUser();
  }

  /**
   * Tries to detect token type from header and returns decoded and verified token
   * 
   * @param header - authorization header value
   * @throws InvalidTokenException - throws if failed to decode token or token is invalid
   * @return verified AccessToken
   */
  @NotNull
  public AccessToken getAccessToken(@Nullable String header) throws InvalidTokenException {
    final AccessToken token;
    if (AccessToken.isBearer(header)) {
      token = getAccessTokenFromBearer(header);
    } else if (BasicAuthToken.isBasic(header)) {
      token = getAccessTokenFromBasic(header);
    } else {
      throw new InvalidAuthorizationHeaderException(AccessToken.BEARER_TYPE + " or " + BasicAuthToken.BASIC_TYPE);
    }
    return verify(token);
  }

  /**
   * Tries to decode Basic Authorization header and issue AccessToken without verification
   * 
   * @param basicAuthorizationHeader - Basic authorization header value
   * @throws InvalidTokenException - throws if failed to decode token or token is invalid
   * @return not verified AccessToken
   */
  @NotNull
  public AccessToken getAccessTokenFromBasic(@NotNull String basicAuthorizationHeader) throws InvalidTokenException {
    BasicAuthToken basic = BasicAuthToken.decodeHeader(basicAuthorizationHeader);
    return client.getTokenHolder().getToken(basic).getAccessToken();
  }

  /**
   * Tries to decode Bearer Authorization header with OAuth2 AccessToken inside
   * 
   * @param bearerAuthorizationHeader - Bearer authorization header value
   * @throws InvalidTokenException - throws if failed to decode token or token is invalid
   * @return not verified AccessToken
   */
  @NotNull
  public AccessToken getAccessTokenFromBearer(@NotNull String bearerAuthorizationHeader) throws InvalidTokenException {
    return AccessToken.decodeHeader(bearerAuthorizationHeader);
  }

  @NotNull
  public AccessToken verify(@NotNull AccessToken token) throws InvalidTokenException {

    if (clientServiceId == null) {
      String clientServiceId = client.getClientServiceId();
      if ((clientServiceId == null || clientServiceId.length() == 0)) {
        throw new IllegalStateException("Unable to verify empty serviceId");
      }
      this.clientServiceId = clientServiceId;
    }

    AccessTokenVerifier verifier = new AccessTokenVerifier(this.clientServiceId, this.keySource);
    verifier.verify(token);
    return token;
  }

  public String getGuestId() {
    if (guestId == null) {
      synchronized (RequestValidator.class) {
        if (guestId == null) {
          guestId = this.client.getGuest(null).getId();
        }
      }
    }
    return guestId;
  }
}
