package jetbrains.jetpass.dao.util;

import java.util.Collections;
import java.util.Iterator;
import java.util.NoSuchElementException;

abstract class AbstractPersistent23Tree<K extends Comparable<K>> implements Iterable<K> {

    abstract RootNode<K> getRoot();

    public boolean contains(K key) {
        Node<K> root = getRoot();
        return root != null && root.get(key) != null;
    }

    public boolean isEmpty() {
        return getRoot() == null;
    }

    public int size() {
        final RootNode<K> root = getRoot();
        return root == null ? 0 : root.getSize();
    }

    public final K getMinimum() {
        final Iterator<K> it = iterator();
        return it.hasNext() ? it.next() : null;
    }

    public final K getMaximum() {
        final Iterator<K> it = reverseIterator();
        return it.hasNext() ? it.next() : null;
    }

    private static class TreePos<K extends Comparable<K>> {

        final Node<K> node;
        int pos;

        TreePos(Node<K> node) {
            this.node = node;
        }
    }

    @Override
    public Iterator<K> iterator() {
        //noinspection unchecked
        return isEmpty() ? Collections.EMPTY_LIST.iterator() : new Iterator<K>() {

            private Stack<TreePos<K>> stack;
            private boolean hasNext;
            private boolean hasNextValid;

            @Override
            public boolean hasNext() {
                if (hasNextValid) {
                    return hasNext;
                }
                hasNextValid = true;
                if (stack == null) {
                    Node<K> root = getRoot();
                    if (root == null) {
                        hasNext = false;
                        return hasNext;
                    }
                    stack = new Stack<TreePos<K>>();
                    stack.push(new TreePos<K>(root));
                }
                TreePos<K> treePos = stack.peek();
                if (treePos.node.isLeaf()) {
                    while (treePos.pos >= (treePos.node.isTernary() ? 2 : 1)) {
                        stack.pop();
                        if (stack.isEmpty()) {
                            hasNext = false;
                            return hasNext;
                        }
                        treePos = stack.peek();
                    }
                } else {
                    if (treePos.pos == 0) {
                        treePos = new TreePos<K>(treePos.node.getFirstChild());
                    } else if (treePos.pos == 1) {
                        treePos = new TreePos<K>(treePos.node.getSecondChild());
                    } else {
                        treePos = new TreePos<K>(treePos.node.getThirdChild());
                    }
                    stack.push(treePos);
                    while (!treePos.node.isLeaf()) {
                        treePos = new TreePos<K>(treePos.node.getFirstChild());
                        stack.push(treePos);
                    }
                }
                treePos.pos++;
                hasNext = true;
                return hasNext;
            }

            @Override
            public K next() {
                if (!hasNext()) {
                    throw new NoSuchElementException();
                }
                hasNextValid = false;
                TreePos<K> treePos = stack.peek();
                // treePos.pos must be 1 or 2 here
                return treePos.pos == 1 ? treePos.node.getFirstKey() : treePos.node.getSecondKey();
            }

            @Override
            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }

    /**
     * Returns an iterator that iterates over all elements greater or equal to key in ascending order
     *
     * @param key key
     * @return iterator
     */
    public Iterator<K> tailIterator(final K key) {
        return new Iterator<K>() {

            private Stack<TreePos<K>> stack;
            private boolean hasNext;
            private boolean hasNextValid;

            @Override
            public boolean hasNext() {
                if (hasNextValid) {
                    return hasNext;
                }
                hasNextValid = true;
                if (stack == null) {
                    Node<K> root = getRoot();
                    if (root == null) {
                        hasNext = false;
                        return hasNext;
                    }
                    stack = new Stack<TreePos<K>>();
                    if (!root.getLess(key, stack)) {
                        stack.push(new TreePos<K>(root));
                    }
                }
                TreePos<K> treePos = stack.peek();
                if (treePos.node.isLeaf()) {
                    while (treePos.pos >= (treePos.node.isTernary() ? 2 : 1)) {
                        stack.pop();
                        if (stack.isEmpty()) {
                            hasNext = false;
                            return hasNext;
                        }
                        treePos = stack.peek();
                    }
                } else {
                    if (treePos.pos == 0) {
                        treePos = new TreePos<K>(treePos.node.getFirstChild());
                    } else if (treePos.pos == 1) {
                        treePos = new TreePos<K>(treePos.node.getSecondChild());
                    } else {
                        treePos = new TreePos<K>(treePos.node.getThirdChild());
                    }
                    stack.push(treePos);
                    while (!treePos.node.isLeaf()) {
                        treePos = new TreePos<K>(treePos.node.getFirstChild());
                        stack.push(treePos);
                    }
                }
                treePos.pos++;
                hasNext = true;
                return hasNext;
            }

            @Override
            public K next() {
                if (!hasNext()) {
                    throw new NoSuchElementException();
                }
                hasNextValid = false;
                TreePos<K> treePos = stack.peek();
                // treePos.pos must be 1 or 2 here
                return treePos.pos == 1 ? treePos.node.getFirstKey() : treePos.node.getSecondKey();
            }

            @Override
            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }

    class TreePosRev<K extends Comparable<K>> {
        private final Node<K> node;
        private int pos;

        TreePosRev(Node<K> node) {
            this.node = node;
            pos = 2;
            if (node.isTernary()) {
                pos = 3;
            }
        }
    }

    public Iterator<K> reverseIterator() {
        return new Iterator<K>() {

            private Stack<TreePosRev<K>> stack;
            private boolean hasNext;
            private boolean hasNextValid;

            @Override
            public boolean hasNext() {
                if (hasNextValid) {
                    return hasNext;
                }
                hasNextValid = true;
                if (stack == null) {
                    Node<K> root = getRoot();
                    if (root == null) {
                        hasNext = false;
                        return hasNext;
                    }
                    stack = new Stack<TreePosRev<K>>();
                    stack.push(new TreePosRev<K>(root));
                }
                TreePosRev<K> treePos = stack.peek();
                if (treePos.node.isLeaf()) {
                    while (treePos.pos <= 1) {
                        stack.pop();
                        if (stack.isEmpty()) {
                            hasNext = false;
                            return hasNext;
                        }
                        treePos = stack.peek();
                    }
                } else {
                    if (treePos.pos == 1) {
                        treePos = new TreePosRev<K>(treePos.node.getFirstChild());
                    } else if (treePos.pos == 2) {
                        treePos = new TreePosRev<K>(treePos.node.getSecondChild());
                    } else {
                        treePos = new TreePosRev<K>(treePos.node.getThirdChild());
                    }
                    stack.push(treePos);
                    while (!treePos.node.isLeaf()) {
                        treePos = new TreePosRev<K>(treePos.node.isTernary() ? treePos.node.getThirdChild() : treePos.node.getSecondChild());
                        stack.push(treePos);
                    }
                }
                treePos.pos--;
                hasNext = true;
                return hasNext;
            }

            @Override
            public K next() {
                if (!hasNext()) {
                    throw new NoSuchElementException();
                }
                hasNextValid = false;
                TreePosRev<K> treePos = stack.peek();
                // treePos.pos must be 1 or 2 here
                return treePos.pos == 1 ? treePos.node.getFirstKey() : treePos.node.getSecondKey();
            }

            @Override
            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }

    /**
     * Returns an iterator that iterates over all elements greater or equal to key in descending order
     *
     * @param key key
     * @return iterator
     */
    public Iterator<K> tailReverseIterator(final K key) {
        return new Iterator<K>() {

            private Stack<TreePosRev<K>> stack;
            private boolean hasNext;
            private boolean hasNextValid;
            private K bound;

            @Override
            public boolean hasNext() {
                if (hasNextValid) {
                    return hasNext;
                }
                hasNextValid = true;
                if (stack == null) {
                    Node<K> root = getRoot();
                    if (root == null) {
                        hasNext = false;
                        return hasNext;
                    }
                    bound = getLess(root, key);
                    stack = new Stack<TreePosRev<K>>();
                    stack.push(new TreePosRev<K>(root));
                }
                TreePosRev<K> treePos = stack.peek();
                if (treePos.node.isLeaf()) {
                    while (treePos.pos <= 1) {
                        stack.pop();
                        if (stack.isEmpty()) {
                            hasNext = false;
                            return hasNext;
                        }
                        treePos = stack.peek();
                    }
                } else {
                    if (treePos.pos == 1) {
                        treePos = new TreePosRev<K>(treePos.node.getFirstChild());
                    } else if (treePos.pos == 2) {
                        treePos = new TreePosRev<K>(treePos.node.getSecondChild());
                    } else {
                        treePos = new TreePosRev<K>(treePos.node.getThirdChild());
                    }
                    stack.push(treePos);
                    while (!treePos.node.isLeaf()) {
                        treePos = new TreePosRev<K>(treePos.node.isTernary() ? treePos.node.getThirdChild() : treePos.node.getSecondChild());
                        stack.push(treePos);
                    }
                }
                treePos.pos--;
                hasNext = tryNext() != bound;
                return hasNext;
            }

            @Override
            public K next() {
                if (!hasNext()) {
                    throw new NoSuchElementException();
                }
                hasNextValid = false;
                return tryNext();
            }

            private K tryNext() {
                TreePosRev<K> treePos = stack.peek();
                // treePos.pos must be 1 or 2 here
                return treePos.pos == 1 ? treePos.node.getFirstKey() : treePos.node.getSecondKey();
            }

            @Override
            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }

    static <K extends Comparable<K>> Node<K> createNode(Node<K> firstChild, K key, Node<K> secondChild) {
        if (firstChild == null && secondChild == null) {
            return new BinaryNode<K>(key);
        } else {
            return new InternalBinaryNode<K>(firstChild, key, secondChild);
        }
    }

    static <K extends Comparable<K>> RootNode<K> createRootNode(Node<K> firstChild, K key,
                                                                Node<K> secondChild, int size) {
        if (firstChild == null && secondChild == null) {
            return new RootBinaryNode<K>(key, size);
        } else {
            return new RootInternalBinaryNode<K>(firstChild, key, secondChild, size);
        }
    }

    static <K extends Comparable<K>> Node<K> createNode(Node<K> firstChild, K firstKey,
                                                        Node<K> secondChild, K secondKey,
                                                        Node<K> thirdChild) {
        if (firstChild == null && secondChild == null && thirdChild == null) {
            return new TernaryNode<K>(firstKey, secondKey);
        } else {
            return new InternalTernaryNode<K>(firstChild, firstKey, secondChild, secondKey, thirdChild);
        }
    }

    static <K extends Comparable<K>> RootNode<K> createRootNode(Node<K> firstChild, K firstKey,
                                                                Node<K> secondChild, K secondKey,
                                                                Node<K> thirdChild, int size) {
        if (firstChild == null && secondChild == null && thirdChild == null) {
            return new RootTernaryNode<K>(firstKey, secondKey, size);
        } else {
            return new RootInternalTernaryNode<K>(firstChild, firstKey, secondChild, secondKey, thirdChild, size);
        }
    }

    interface Node<K extends Comparable<K>> {
        Node<K> getFirstChild();

        Node<K> getSecondChild();

        Node<K> getThirdChild();

        K getFirstKey();

        K getSecondKey();

        boolean isLeaf();

        boolean isTernary();

        RootNode<K> asRoot(int size);

        /**
         * Inserts key into tree.
         * If tree already contains K k for which key.compareTo(k) == 0 tree *is* modified.
         * This is used by implementation of map using this set.
         *
         * @param key key to insert
         * @return result of insertion: either resulting node or two nodes with key to be added on the parent level
         */

        SplitResult<K> insert(K key);

        /**
         * @param key    key to remove
         * @param strict if strict is true removes key, else removes the first greater key
         * @return Replacing node of the new version of the tree and the removed key.
         * Node will be null if key's not found, replacing node of the resulting tree if no merge required,
         * node with single child and no key if merge is needed.
         * Key will be null if there's no such key.
         */
        Pair<Node<K>, K> remove(K key, boolean strict);

        K get(K key);

        boolean getLess(K key, Stack<TreePos<K>> stack);

        /**
         * checks the subtree of this node
         *
         * @param from every key in this subtree must be at least from
         * @param to   every key in this subtree must be at most to
         * @return the depth of this subtree: must be the same for all the children
         */
        int checkNode(K from, K to);

        String print();

        void count(int[] count);
    }

    interface RootNode<K extends Comparable<K>> extends Node<K> {

        int getSize();
    }

    /**
     * checks the subtree of the node for correctness: for tests only
     *
     * @throws RuntimeException if the tree structure is incorrect
     */
    static <K extends Comparable<K>> void checkNode(Node<K> node) {
        node.checkNode(null, null);
    }

    static <K extends Comparable<K>> boolean getLess(Node<K> node, Stack<TreePos<K>> stack) {
        stack.push(new TreePos<K>(node));
        return false;
    }

    static <K extends Comparable<K>> K getLess(Node<K> node, K key) {
        Stack<TreePos<K>> stack = new Stack<TreePos<K>>();
        if (!node.getLess(key, stack)) {
            return null;
        }
        TreePos<K> treePos = stack.peek();
        return treePos.pos == 1 ? treePos.node.getFirstKey() : treePos.node.getSecondKey();
    }

    static class RootBinaryNode<K extends Comparable<K>> extends BinaryNode<K> implements RootNode<K> {

        private final int size;

        RootBinaryNode(K firstKey, int size) {
            super(firstKey);
            this.size = size;
        }

        @Override
        public int getSize() {
            return size;
        }
    }

    static class BinaryNode<K extends Comparable<K>> implements Node<K> {

        private final K firstKey;

        BinaryNode(K firstKey) {
            this.firstKey = firstKey;
        }

        @Override
        public Node<K> getFirstChild() {
            return null;
        }

        @Override
        public Node<K> getSecondChild() {
            return null;
        }

        @Override
        public Node<K> getThirdChild() {
            throw new UnsupportedOperationException();
        }

        @Override
        public K getFirstKey() {
            return firstKey;
        }

        @Override
        public K getSecondKey() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean isLeaf() {
            return true;
        }

        @Override
        public boolean isTernary() {
            return false;
        }

        @Override
        public RootNode<K> asRoot(int size) {
            return new RootBinaryNode<K>(firstKey, size);
        }

        @Override
        public SplitResult<K> insert(K key) {
            int comp = key.compareTo(firstKey);
            if (comp < 0) {
                return new SplitResult<K>().fill(new TernaryNode<K>(key, firstKey)).setSizeChanged();
            }
            if (comp == 0) {
                return new SplitResult<K>().fill(new BinaryNode<K>(key));
            }
            return new SplitResult<K>().fill(new TernaryNode<K>(firstKey, key)).setSizeChanged();
        }

        @Override
        public Pair<Node<K>, K> remove(K key, boolean strict) {
            int comp = strict ? key.compareTo(firstKey) : -1;
            if (strict && comp != 0) {
                return null;
            } else {
                return new Pair<Node<K>, K>(new RemovedNode<K>(null), firstKey);
            }
        }

        @Override
        public K get(K key) {
            return key.compareTo(firstKey) == 0 ? firstKey : null;
        }

        @Override
        public boolean getLess(K key, Stack<TreePos<K>> stack) {
            AbstractPersistent23Tree.getLess(this, stack);
            if (firstKey.compareTo(key) >= 0) {
                stack.pop();
                return false;
            }
            stack.peek().pos++;
            return true;
        }

        @Override
        public int checkNode(K from, K to) {
            if (from != null && from.compareTo(firstKey) >= 0) {
                throw new RuntimeException("Not a search tree.");
            }
            if (to != null && to.compareTo(firstKey) <= 0) {
                throw new RuntimeException("Not a search tree.");
            }
            return 1;
        }

        @Override
        public String print() {
            return String.valueOf(firstKey);
        }

        @Override
        public void count(int[] count) {
            count[0]++;
        }
    }

    static class RootInternalBinaryNode<K extends Comparable<K>> extends InternalBinaryNode<K> implements RootNode<K> {

        private final int size;

        RootInternalBinaryNode(Node<K> firstChild, K firstKey, Node<K> secondChild, int size) {
            super(firstChild, firstKey, secondChild);
            this.size = size;
        }

        @Override
        public int getSize() {
            return size;
        }
    }

    static class InternalBinaryNode<K extends Comparable<K>> implements Node<K> {

        private final K firstKey;
        private final Node<K> firstChild;
        private final Node<K> secondChild;

        InternalBinaryNode(Node<K> firstChild, K firstKey, Node<K> secondChild) {
            this.firstKey = firstKey;
            this.firstChild = firstChild;
            this.secondChild = secondChild;
        }

        @Override
        public Node<K> getFirstChild() {
            return firstChild;
        }

        @Override
        public Node<K> getSecondChild() {
            return secondChild;
        }

        @Override
        public Node<K> getThirdChild() {
            throw new UnsupportedOperationException();
        }

        @Override
        public K getFirstKey() {
            return firstKey;
        }

        @Override
        public K getSecondKey() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean isLeaf() {
            return false;
        }

        @Override
        public boolean isTernary() {
            return false;
        }

        @Override
        public RootNode<K> asRoot(int size) {
            return new RootInternalBinaryNode<K>(firstChild, firstKey, secondChild, size);
        }


        @Override
        public SplitResult<K> insert(K key) {
            int comp = key.compareTo(firstKey);
            if (comp < 0) {
                SplitResult<K> splitResult = firstChild.insert(key);
                final Node<K> firstNode = splitResult.getFirstNode();
                final K splitKey = splitResult.getKey();
                // no split
                if (splitKey == null) {
                    return splitResult.fill(new InternalBinaryNode<K>(firstNode, firstKey, secondChild));
                }
                // split occurred
                return splitResult.fill(new InternalTernaryNode<K>(firstNode, splitKey, splitResult.getSecondNode(), firstKey, secondChild));
            }
            if (comp == 0) {
                return new SplitResult<K>().fill(new InternalBinaryNode<K>(firstChild, key, secondChild));
            }

            SplitResult<K> splitResult = secondChild.insert(key);
            final Node<K> firstNode = splitResult.getFirstNode();
            final K splitKey = splitResult.getKey();
            // no split
            if (splitKey == null) {
                return splitResult.fill(new InternalBinaryNode<K>(firstChild, firstKey, firstNode));
            }
            // split occurred
            return splitResult.fill(new InternalTernaryNode<K>(firstChild, firstKey, firstNode, splitKey, splitResult.getSecondNode()));
        }

        @SuppressWarnings({"SimplifiableConditionalExpression", "ConditionalExpressionWithNegatedCondition", "OverlyLongMethod"})
        @Override
        public Pair<Node<K>, K> remove(K key, boolean strict) {
            int comp = strict ? key.compareTo(firstKey) : -1;
            if (comp < 0) {
                Pair<Node<K>, K> removeResult = firstChild.remove(key, strict);
                if (removeResult == null) {
                    return null;
                }
                Node<K> resultNode = removeResult.getFirst();
                K removedKey = removeResult.getSecond();
                if (resultNode instanceof RemovedNode) {
                    Node<K> removedNodeResult = resultNode.getFirstChild();
                    if (!secondChild.isTernary()) {
                        final Node<K> node = createNode(
                                removedNodeResult, firstKey, secondChild.getFirstChild(), secondChild.getFirstKey(), secondChild.getSecondChild()
                        );
                        return new Pair<Node<K>, K>(new RemovedNode<K>(node), removedKey);
                    } else {
                        return new Pair<Node<K>, K>(new InternalBinaryNode<K>(
                                createNode(removedNodeResult, firstKey, secondChild.getFirstChild()), secondChild.getFirstKey(),
                                createNode(secondChild.getSecondChild(), secondChild.getSecondKey(), secondChild.getThirdChild())), removedKey);
                    }
                } else {
                    return new Pair<Node<K>, K>(createNode(resultNode, firstKey, secondChild), removedKey);
                }
            }
            // strict is true here
            Pair<Node<K>, K> removeResult = secondChild.remove(key, comp != 0);
            if (removeResult == null) {
                return null;
            }
            Node<K> resultNode = removeResult.getFirst();
            K removedKey = comp == 0 ? firstKey : removeResult.getSecond();
            K newNodeKey = comp != 0 ? firstKey : removeResult.getSecond();
            if (resultNode instanceof RemovedNode) {
                Node<K> removedNodeResult = resultNode.getFirstChild();
                if (!firstChild.isTernary()) {
                    return new Pair<Node<K>, K>(new RemovedNode<K>(createNode(
                            firstChild.getFirstChild(), firstChild.getFirstKey(), firstChild.getSecondChild(), newNodeKey, removedNodeResult
                    )), removedKey);
                } else {
                    return new Pair<Node<K>, K>(new InternalBinaryNode<K>(
                            createNode(firstChild.getFirstChild(), firstChild.getFirstKey(), firstChild.getSecondChild()),
                            firstChild.getSecondKey(), createNode(firstChild.getThirdChild(), newNodeKey, removedNodeResult)), removedKey);
                }
            } else {
                return new Pair<Node<K>, K>(new InternalBinaryNode<K>(firstChild, newNodeKey, resultNode), removedKey);
            }
        }

        @Override
        public K get(K key) {
            int comp = key.compareTo(firstKey);
            if (comp == 0) {
                return firstKey;
            }
            if (comp < 0) {
                return firstChild.get(key);
            } else {
                return secondChild.get(key);
            }
        }

        @Override
        public boolean getLess(K key, Stack<TreePos<K>> stack) {
            AbstractPersistent23Tree.getLess(this, stack);
            int comp = firstKey.compareTo(key);
            if (comp < 0) {
                stack.peek().pos++;
                secondChild.getLess(key, stack);
            } else if (!firstChild.getLess(key, stack)) {
                stack.pop();
                return false;
            }
            return true;
        }

        @Override
        public int checkNode(K from, K to) {
            if (from != null && from.compareTo(firstKey) >= 0) {
                throw new RuntimeException("Not a search tree.");
            }
            if (to != null && to.compareTo(firstKey) <= 0) {
                throw new RuntimeException("Not a search tree.");
            }
            if (firstChild == null || secondChild == null) {
                throw new RuntimeException("Not an inner node.");
            }
            int firstDepth = firstChild.checkNode(from, firstKey);
            int secondDepth = secondChild.checkNode(firstKey, to);
            if (firstDepth != secondDepth) {
                throw new RuntimeException("Not balanced tree.");
            }
            return firstDepth + 1;
        }

        @Override
        public String print() {
            return '(' + firstChild.print() + ") " + firstKey + " (" + secondChild.print() + ')';
        }

        @Override
        public void count(int[] count) {
            count[1]++;
            firstChild.count(count);
            secondChild.count(count);
        }
    }

    static class RootTernaryNode<K extends Comparable<K>> extends TernaryNode<K> implements RootNode<K> {

        private final int size;

        RootTernaryNode(K firstKey, K secondKey, int size) {
            super(firstKey, secondKey);
            this.size = size;
        }

        @Override
        public int getSize() {
            return size;
        }
    }

    static class TernaryNode<K extends Comparable<K>> implements Node<K> {

        private final K firstKey;
        private final K secondKey;

        TernaryNode(K firstKey, K secondKey) {
            this.firstKey = firstKey;
            this.secondKey = secondKey;
        }

        @Override
        public Node<K> getFirstChild() {
            return null;
        }

        @Override
        public Node<K> getSecondChild() {
            return null;
        }

        @Override
        public Node<K> getThirdChild() {
            return null;
        }

        @Override
        public K getFirstKey() {
            return firstKey;
        }

        @Override
        public K getSecondKey() {
            return secondKey;
        }

        @Override
        public boolean isLeaf() {
            return true;
        }

        @Override
        public boolean isTernary() {
            return true;
        }

        @Override
        public RootNode<K> asRoot(int size) {
            return new RootTernaryNode<K>(firstKey, secondKey, size);
        }

        @Override
        public SplitResult<K> insert(K key) {
            int comp = key.compareTo(firstKey);
            if (comp < 0) {
                return new SplitResult<K>().fill(new BinaryNode<K>(key), firstKey, new BinaryNode<K>(secondKey)).setSizeChanged();
            }
            if (comp == 0) {
                return new SplitResult<K>().fill(new TernaryNode<K>(key, secondKey));
            }
            comp = key.compareTo(secondKey);
            if (comp < 0) {
                return new SplitResult<K>().fill(new BinaryNode<K>(firstKey), key, new BinaryNode<K>(secondKey)).setSizeChanged();
            }
            if (comp == 0) {
                return new SplitResult<K>().fill(new TernaryNode<K>(firstKey, key));
            }
            return new SplitResult<K>().fill(new BinaryNode<K>(firstKey), secondKey, new BinaryNode<K>(key)).setSizeChanged();
        }

        @Override
        public Pair<Node<K>, K> remove(K key, boolean strict) {
            int compFirst = strict ? key.compareTo(firstKey) : -1;
            if (compFirst < 0 && strict) {
                return null;
            }
            if (compFirst <= 0) {
                return new Pair<Node<K>, K>(new BinaryNode<K>(secondKey), firstKey);
            }
            // strict is true here
            int compSecond = -1;
            if (compFirst > 0) {
                compSecond = key.compareTo(secondKey);
            }
            if (compSecond != 0) {
                return null;
            } else {
                return new Pair<Node<K>, K>(new BinaryNode<K>(firstKey), secondKey);
            }
        }

        @Override
        public K get(K key) {
            int comp = key.compareTo(firstKey);
            if (comp == 0) {
                return firstKey;
            }
            if (comp > 0 && key.compareTo(secondKey) == 0) {
                return secondKey;
            }
            return null;
        }

        @Override
        public boolean getLess(K key, Stack<TreePos<K>> stack) {
            AbstractPersistent23Tree.getLess(this, stack);
            if (firstKey.compareTo(key) >= 0) {
                stack.pop();
                return false;
            }
            if (secondKey.compareTo(key) < 0) {
                stack.peek().pos += 2;
            } else {
                stack.peek().pos++;
            }
            return true;
        }

        @Override
        public int checkNode(K from, K to) {
            if (from != null && from.compareTo(firstKey) >= 0) {
                throw new RuntimeException("Not a search tree.");
            }
            if (firstKey.compareTo(secondKey) >= 0) {
                throw new RuntimeException("Not a search tree.");
            }
            if (to != null && to.compareTo(secondKey) <= 0) {
                throw new RuntimeException("Not a search tree.");
            }
            return 1;
        }

        @Override
        public String print() {
            return firstKey + ", " + secondKey;
        }

        @Override
        public void count(int[] count) {
            count[2]++;
        }
    }

    static class RootInternalTernaryNode<K extends Comparable<K>> extends InternalTernaryNode<K> implements RootNode<K> {

        private final int size;

        RootInternalTernaryNode(Node<K> firstChild, K firstKey,
                                Node<K> secondChild, K secondKey,
                                Node<K> thirdChild, int size) {
            super(firstChild, firstKey, secondChild, secondKey, thirdChild);
            this.size = size;
        }

        @Override
        public int getSize() {
            return size;
        }
    }

    static class InternalTernaryNode<K extends Comparable<K>> implements Node<K> {

        private final K firstKey;
        private final K secondKey;
        private final Node<K> firstChild;
        private final Node<K> secondChild;
        private final Node<K> thirdChild;

        InternalTernaryNode(Node<K> firstChild, K firstKey, Node<K> secondChild, K secondKey, Node<K> thirdChild) {
            this.firstKey = firstKey;
            this.secondKey = secondKey;
            this.firstChild = firstChild;
            this.secondChild = secondChild;
            this.thirdChild = thirdChild;
        }

        @Override
        public Node<K> getFirstChild() {
            return firstChild;
        }

        @Override
        public Node<K> getSecondChild() {
            return secondChild;
        }

        @Override
        public Node<K> getThirdChild() {
            return thirdChild;
        }

        @Override
        public K getFirstKey() {
            return firstKey;
        }

        @Override
        public K getSecondKey() {
            return secondKey;
        }

        @Override
        public boolean isLeaf() {
            return false;
        }

        @Override
        public boolean isTernary() {
            return true;
        }

        @Override
        public RootNode<K> asRoot(int size) {
            return new RootInternalTernaryNode<K>(firstChild, firstKey, secondChild, secondKey, thirdChild, size);
        }


        @Override
        public SplitResult<K> insert(K key) {
            int comp = key.compareTo(firstKey);
            if (comp < 0) {
                SplitResult<K> splitResult = firstChild.insert(key);
                final Node<K> firstNode = splitResult.getFirstNode();
                final K splitKey = splitResult.getKey();
                // no split
                if (splitKey == null) {
                    return splitResult.fill(cloneReplacingChild(firstChild, firstNode));
                }
                // split occurred
                return splitResult.fill(new InternalBinaryNode<K>(firstNode, splitKey, splitResult.getSecondNode()), firstKey, new InternalBinaryNode<K>(secondChild, secondKey, thirdChild));
            }
            if (comp == 0) {
                return new SplitResult<K>().fill(new InternalTernaryNode<K>(firstChild, key, secondChild, secondKey, thirdChild));
            }

            comp = key.compareTo(secondKey);
            if (comp < 0) {
                SplitResult<K> splitResult = secondChild.insert(key);
                final Node<K> firstNode = splitResult.getFirstNode();
                final K splitKey = splitResult.getKey();
                // no split
                if (splitKey == null) {
                    return splitResult.fill(cloneReplacingChild(secondChild, firstNode));
                }
                // split occurred
                return splitResult.fill(new InternalBinaryNode<K>(firstChild, firstKey, firstNode), splitKey, new InternalBinaryNode<K>(splitResult.getSecondNode(), secondKey, thirdChild));
            }
            if (comp == 0) {
                return new SplitResult<K>().fill(new InternalTernaryNode<K>(firstChild, firstKey, secondChild, key, thirdChild));
            }

            // node.secondKey != null
            SplitResult<K> splitResult = thirdChild.insert(key);
            final Node<K> firstNode = splitResult.getFirstNode();
            final K splitKey = splitResult.getKey();
            // no split
            if (splitKey == null) {
                return splitResult.fill(cloneReplacingChild(thirdChild, firstNode));
            }
            // split occurred
            return splitResult.fill(new InternalBinaryNode<K>(firstChild, firstKey, secondChild), secondKey, new InternalBinaryNode<K>(firstNode, splitKey, splitResult.getSecondNode()));
        }

        @SuppressWarnings({"ConditionalExpressionWithNegatedCondition", "OverlyLongMethod"})
        @Override
        public Pair<Node<K>, K> remove(K key, boolean strict) {
            int compFirst = strict ? key.compareTo(firstKey) : -1;
            if (compFirst < 0) {
                Pair<Node<K>, K> removeResult = firstChild.remove(key, strict);
                if (removeResult == null) {
                    return null;
                }
                Node<K> resultNode = removeResult.getFirst();
                K removedKey = removeResult.getSecond();
                if (resultNode instanceof RemovedNode) {
                    Node<K> removedNodeResult = resultNode.getFirstChild();
                    if (!secondChild.isTernary()) {
                        return new Pair<Node<K>, K>(new InternalBinaryNode<K>(createNode(
                                removedNodeResult, firstKey, secondChild.getFirstChild(), secondChild.getFirstKey(), secondChild.getSecondChild()
                        ), secondKey, thirdChild), removedKey);
                    } else {
                        return new Pair<Node<K>, K>(new InternalTernaryNode<K>(createNode(removedNodeResult, firstKey, secondChild.getFirstChild()),
                                secondChild.getFirstKey(), createNode(secondChild.getSecondChild(), secondChild.getSecondKey(), secondChild.getThirdChild()),
                                secondKey, thirdChild), removedKey);
                    }
                } else {
                    return new Pair<Node<K>, K>(cloneReplacingChild(firstChild, resultNode), removedKey);
                }
            }
            // strict is true here
            int compSecond = -1;
            if (compFirst > 0) {
                compSecond = key.compareTo(secondKey);
            }
            if (compSecond < 0) {
                Pair<Node<K>, K> removeResult = secondChild.remove(key, compFirst != 0);
                if (removeResult == null) {
                    return null;
                }
                Node<K> resultNode = removeResult.getFirst();
                K removedKey = compFirst == 0 ? firstKey : removeResult.getSecond();
                K newNodeKey = compFirst != 0 ? firstKey : removeResult.getSecond();
                if (resultNode instanceof RemovedNode) {
                    Node<K> removedNodeResult = resultNode.getFirstChild();
                    if (!firstChild.isTernary()) {
                        return new Pair<Node<K>, K>(new InternalBinaryNode<K>(createNode(
                                firstChild.getFirstChild(), firstChild.getFirstKey(), firstChild.getSecondChild(), newNodeKey, removedNodeResult
                        ), secondKey, thirdChild), removedKey);
                    } else {
                        return new Pair<Node<K>, K>(new InternalTernaryNode<K>(
                                createNode(firstChild.getFirstChild(), firstChild.getFirstKey(), firstChild.getSecondChild()),
                                firstChild.getSecondKey(), createNode(firstChild.getThirdChild(), newNodeKey, removedNodeResult), secondKey, thirdChild), removedKey);
                    }
                } else {
                    return new Pair<Node<K>, K>(new InternalTernaryNode<K>(firstChild, newNodeKey, resultNode, secondKey, thirdChild), removedKey);
                }
            }

            Pair<Node<K>, K> removeResult = thirdChild.remove(key, compSecond != 0);
            if (removeResult == null) {
                return null;
            }
            Node<K> resultNode = removeResult.getFirst();
            K removedKey = compSecond == 0 ? secondKey : removeResult.getSecond();
            K newNodeKey = compSecond != 0 ? secondKey : removeResult.getSecond();
            if (resultNode instanceof RemovedNode) {
                Node<K> removedNodeResult = resultNode.getFirstChild();
                if (!secondChild.isTernary()) {
                    return new Pair<Node<K>, K>(new InternalBinaryNode<K>(firstChild, firstKey, createNode(
                            secondChild.getFirstChild(), secondChild.getFirstKey(), secondChild.getSecondChild(), newNodeKey, removedNodeResult
                    )), removedKey);
                } else {
                    return new Pair<Node<K>, K>(new InternalTernaryNode<K>(firstChild, firstKey,
                            createNode(secondChild.getFirstChild(), secondChild.getFirstKey(), secondChild.getSecondChild()),
                            secondChild.getSecondKey(), createNode(secondChild.getThirdChild(), newNodeKey, removedNodeResult)), removedKey);
                }
            } else {
                return new Pair<Node<K>, K>(new InternalTernaryNode<K>(firstChild, firstKey, secondChild, newNodeKey, resultNode), removedKey);
            }
        }

        @Override
        public K get(K key) {
            int comp = key.compareTo(firstKey);
            if (comp < 0) {
                return firstChild.get(key);
            }
            if (comp == 0) {
                return firstKey;
            }
            comp = key.compareTo(secondKey);
            if (comp == 0) {
                return secondKey;
            } else {
                return comp < 0 ? secondChild.get(key) : thirdChild.get(key);
            }
        }

        @Override
        public boolean getLess(K key, Stack<TreePos<K>> stack) {
            AbstractPersistent23Tree.getLess(this, stack);
            int comp = secondKey.compareTo(key);
            if (comp < 0) {
                stack.peek().pos += 2;
                thirdChild.getLess(key, stack);
                return true;
            }
            comp = firstKey.compareTo(key);
            if (comp < 0) {
                stack.peek().pos++;
                secondChild.getLess(key, stack);
            } else if (!firstChild.getLess(key, stack)) {
                stack.pop();
                return false;
            }
            return true;
        }

        Node<K> cloneReplacingChild(Node<K> oldChild, Node<K> newChild) {
            return new InternalTernaryNode<K>(firstChild == oldChild ? newChild : firstChild, firstKey,
                    secondChild == oldChild ? newChild : secondChild, secondKey,
                    thirdChild == oldChild ? newChild : thirdChild);
        }

        @Override
        public int checkNode(K from, K to) {
            if (from != null && from.compareTo(firstKey) >= 0) {
                throw new RuntimeException("Not a search tree.");
            }
            if (firstKey.compareTo(secondKey) >= 0) {
                throw new RuntimeException("Not a search tree.");
            }
            if (to != null && to.compareTo(secondKey) <= 0) {
                throw new RuntimeException("Not a search tree.");
            }
            if (firstChild == null || secondChild == null || thirdChild == null) {
                throw new RuntimeException("The node has not enough children.");
            }
            int depth = firstChild.checkNode(from, firstKey);
            if (depth != secondChild.checkNode(firstKey, secondKey) ||
                    depth != thirdChild.checkNode(secondKey, to)) {
                throw new RuntimeException("Not a balanced tree.");
            }
            return depth + 1;
        }

        @Override
        public String print() {
            return '(' + firstChild.print() + ") " + firstKey + " (" + secondChild.print() + ") " + secondKey + " (" + thirdChild.print() + ')';
        }

        @Override
        public void count(int[] count) {
            count[3]++;
            firstChild.count(count);
            secondChild.count(count);
            thirdChild.count(count);
        }
    }

    static class RemovedNode<K extends Comparable<K>> implements Node<K> {

        private final Node<K> child;

        RemovedNode(Node<K> child) {
            this.child = child;
        }

        @Override
        public Node<K> getFirstChild() {
            return child;
        }

        @Override
        public Node<K> getSecondChild() {
            throw new UnsupportedOperationException();
        }

        @Override
        public Node<K> getThirdChild() {
            throw new UnsupportedOperationException();
        }

        @Override
        public K getFirstKey() {
            throw new UnsupportedOperationException();
        }

        @Override
        public K getSecondKey() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean isLeaf() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean isTernary() {
            throw new UnsupportedOperationException();
        }

        @Override
        public RootNode<K> asRoot(int size) {
            throw new UnsupportedOperationException();
        }


        @Override
        public SplitResult<K> insert(K key) {
            throw new UnsupportedOperationException("Can't insert into a temporary tree.");
        }

        @Override
        public Pair<Node<K>, K> remove(K key, boolean strict) {
            throw new UnsupportedOperationException("Can't remove from a temporary tree.");
        }

        @Override
        public K get(K key) {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean getLess(K key, Stack<TreePos<K>> stack) {
            throw new UnsupportedOperationException();
        }

        @Override
        public int checkNode(K from, K to) {
            throw new UnsupportedOperationException();
        }

        @Override
        public String print() {
            throw new UnsupportedOperationException();
        }

        @Override
        public void count(int[] count) {
            throw new UnsupportedOperationException();
        }
    }

    /**
     * Used for adding values. Can be in one of the following states:
     * 1) Only firstNode is not null.
     * Means there was no split.
     * 2) All fields are not null.
     * Means the key should be inserted into the parent with firstNode and secondNode children instead the old single child.
     */
    static class SplitResult<K extends Comparable<K>> {

        private Node<K> firstNode;
        private Node<K> secondNode;
        private K key;
        boolean sizeChanged = false;

        SplitResult<K> fill(Node<K> first, K key, Node<K> second) {
            firstNode = first;
            this.key = key;
            secondNode = second;
            return this;
        }

        public SplitResult<K> fill(Node<K> node) {
            return fill(node, null, null);
        }

        public Node<K> getFirstNode() {
            return firstNode;
        }

        public Node<K> getSecondNode() {
            return secondNode;
        }

        public K getKey() {
            return key;
        }

        public SplitResult<K> setSizeChanged() {
            sizeChanged = true;
            return this;
        }
    }
}
