package jetbrains.jetpass.dao.util;

public class PersistentHashSet<K> {

    /**
     * The root of the last version of the tree.
     */
    
    private AbstractPersistentHashSet.RootTableNode<K> root;

    @SuppressWarnings("unchecked")
    public PersistentHashSet() {
        this(AbstractPersistentHashSet.EMPTY_ROOT);
    }

    public PersistentHashSet( final AbstractPersistentHashSet.RootTableNode<K> root) {
        this.root = root;
    }

    public ImmutablePersistentHashSet<K> getCurrent() {
        return new ImmutablePersistentHashSet<K>(root);
    }

    public PersistentHashSet<K> getClone() {
        return new PersistentHashSet<K>(root);
    }

    public MutablePersistentHashSet<K> beginWrite() {
        return new MutablePersistentHashSet<K>(this);
    }

    boolean endWrite(MutablePersistentHashSet<K> tree) {
        if (root != tree.getStartingRoot()) {
            return false;
        }
        root = tree.getRoot();
        return true;
    }

    
    AbstractPersistentHashSet.RootTableNode<K> getRoot() {
        return root;
    }

    public static class ImmutablePersistentHashSet<K> extends AbstractPersistentHashSet<K> {

        
        private final RootTableNode<K> root;

        ImmutablePersistentHashSet( final RootTableNode<K> root) {
            this.root = root;
        }

        @Override
        
        RootTableNode<K> getRoot() {
            return root;
        }
    }

    public static class MutablePersistentHashSet<K> extends AbstractPersistentHashSet<K> implements Flag {

        private final PersistentHashSet<K> baseTree;
        private RootTableNode<K> startingRoot;
        private RootTableNode<K> root;
        private boolean flagged;

        MutablePersistentHashSet( PersistentHashSet<K> tree) {
            startingRoot = tree.getRoot();
            root = startingRoot;
            baseTree = tree;
        }

        @Override
        public boolean flag() {
            return flagged = true;
        }

        public void add( K key) {
            RootTableNode<K> actualRoot = root;
            flagged = false;
            Node<K> result = actualRoot.insert(key, key.hashCode(), 0, this);
            root = result.asRoot(flagged ? actualRoot.getSize() + 1 : actualRoot.getSize());
        }

        public boolean remove( K key) {
            RootTableNode<K> actualRoot = root;
            @SuppressWarnings("unchecked") Node<K> result = (Node<K>) actualRoot.remove(key, key.hashCode(), 0);
            if (result != null) {
                root = result.asRoot(actualRoot.getSize() - 1);
                return true;
            }
            return false;
        }

        /**
         * Try to merge changes into the base tree.
         *
         * @return true if merging succeeded
         */
        public boolean endWrite() {
            if (baseTree.endWrite(this)) {
                startingRoot = root;
                return true;
            }
            return false;
        }

        @Override
        
        RootTableNode<K> getRoot() {
            return root;
        }

        TableNode<K> getStartingRoot() {
            return startingRoot;
        }

        public void checkTip() {
            if (root != null) {
                root.checkNode(0);
            }
        }

        public void forEachKey(final ObjectProcedure<K> procedure) {
            root.forEachKey(procedure);
        }
    }
}
